/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using System;
using System.Collections;
using System.Data;
using System.Reflection;
using TopCoder.Data.ConnectionFactory;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class provides a DB specific implementation of the ICollectionIndexSource interface. It holds properties
    /// for connection parameters, either passed in or loaded from the configuration manager. It only works with the
    /// Document table in the database, for loading in document indices.
    /// </summary>
    /// <remarks>
    /// This class is mutable.
    /// </remarks>
    public class DBCollectionIndexSource : ICollectionIndexSource, IDisposable
    {
        /// <summary>
        /// This variable holds the configuration information for this class, including the name of the connection
        /// loaded. This value is initialized to a new <see cref="DBPersistenceParameters"/> instance in the
        /// constructor, using values read from the namespace given to the constructor.
        /// </summary>
        private DBPersistenceParameters parameters;

        /// <summary>
        /// This connection is the actual connection that is used to talk to the database for updates. This value is
        /// set using the Connection Factory component.
        /// </summary>
        private IDbConnection connection = null;

        /// <summary>
        /// A lazy-initialized property for the current connection.
        /// </summary>
        /// <exception cref="Exception">Any exception may be thrown. Consumers need to wrap them.</exception>
        /// <value>
        /// An <see cref="IDbConnection"/>, never null.
        /// </value>
        private IDbConnection Connection
        {
            get
            {
                if (connection == null)
                {
                    // Ask the connection factory
                    connection = ConnectionManager.Instance.CreatePredefinedDbConnection(parameters.ConnectionName);
                    connection.Open();

                    // Start a transaction
                    if (transaction != null)
                    {
                        // Close it, but never fail
                        try
                        {
                            transaction.Dispose();
                        }
                        catch
                        {
                        }
                    }

                    transaction = connection.BeginTransaction();
                }

                return connection;
            }
        }

        /// <summary>
        /// All updates are added to this transaction, and they are committed in the <see cref="Save"/> method. This is
        /// initialized to a new <see cref="IDbTransaction"/> implementation using
        /// <see cref="IDbConnection.BeginTransaction"/>.
        /// </summary>
        private IDbTransaction transaction = null;

        /// <summary>
        /// The command used to load all documents.
        /// </summary>
        private IDbCommand loadDocuments;

        /// <summary>
        /// The command used to load all documents. Managed to always point to the current connection and transaction.
        /// </summary>
        /// <value>
        /// <see cref="loadDocuments"/>
        /// </value>
        private IDbCommand LoadDocuments
        {
            get
            {
                if (loadDocuments == null || loadDocuments.Connection != Connection)
                {
                    if (loadDocuments != null)
                    {
                        // Close, but never fail
                        try
                        {
                            loadDocuments.Dispose();
                        }
                        catch (Exception)
                        {
                        }
                    }

                    loadDocuments = Connection.CreateCommand();
                    loadDocuments.CommandText = "SELECT document_id, document_name FROM Document";

                    // Preparing is often unneeded in modern systems, but it should not hurt, we're doing this once
                    // per connection only.
                    loadDocuments.Prepare();
                }

                if (loadDocuments.Transaction != transaction)
                {
                    loadDocuments.Transaction = transaction;
                }

                return loadDocuments;
            }
        }

        /// <summary>
        /// The command used to add a document.
        /// </summary>
        private IDbCommand addDocument;

        /// <summary>
        /// The command used to add a document. Managed to always point to the current connection and transaction.
        /// </summary>
        /// <value>
        /// <see cref="addDocument"/>
        /// </value>
        private IDbCommand AddDocument
        {
            get
            {
                if (addDocument == null || addDocument.Connection != Connection)
                {
                    if (addDocument != null)
                    {
                        // Close, but never fail
                        try
                        {
                            addDocument.Dispose();
                        }
                        catch (Exception)
                        {
                        }
                    }

                    addDocument = Connection.CreateCommand();
                    addDocument.CommandText = "INSERT INTO Document " +
                        "(document_id, document_name, culture, compare_options) VALUES (?, ?, 'en-US', 'None')";

                    IDbDataParameter document_id = addDocument.CreateParameter();
                    addDocument.Parameters.Add(document_id);

                    IDbDataParameter document_name = addDocument.CreateParameter();
                    addDocument.Parameters.Add(document_name);

                    // Preparing is often unneeded in modern systems, but it should not hurt, we're doing this once
                    // per connection only.
                    addDocument.Prepare();
                }

                if (addDocument.Transaction != transaction)
                {
                    addDocument.Transaction = transaction;
                }

                return addDocument;
            }
        }

        /// <summary>
        /// The command used to remove a document.
        /// </summary>
        private IDbCommand removeDocument;

        /// <summary>
        /// The command used to remove a document. Managed to always point to the current connection and transaction.
        /// </summary>
        /// <value>
        /// <see cref="removeDocument"/>
        /// </value>
        private IDbCommand RemoveDocument
        {
            get
            {
                if (removeDocument == null || removeDocument.Connection != Connection)
                {
                    if (removeDocument != null)
                    {
                        // Close, but never fail
                        try
                        {
                            removeDocument.Dispose();
                        }
                        catch (Exception)
                        {
                        }
                    }

                    removeDocument = Connection.CreateCommand();
                    removeDocument.CommandText = "DELETE FROM Document WHERE document_id = ? AND document_name = ?";

                    // We remove with constraints on both properties, to avoid removing a whole lot incorrectly
                    IDbDataParameter document_id = removeDocument.CreateParameter();
                    removeDocument.Parameters.Add(document_id);

                    IDbDataParameter document_name = removeDocument.CreateParameter();
                    removeDocument.Parameters.Add(document_name);

                    // Preparing is often unneeded in modern systems, but it should not hurt, we're doing this once
                    // per connection only.
                    removeDocument.Prepare();
                }

                if (removeDocument.Transaction != transaction)
                {
                    removeDocument.Transaction = transaction;
                }

                return removeDocument;
            }
        }

        /// <summary>
        /// The command used to remove all documents, words and locations.
        /// </summary>
        private IDbCommand clearCommand;

        /// <summary>
        /// The command used to remove all documents, words and locations.
        /// Managed to always point to the current connection and transaction.
        /// </summary>
        /// <value>
        /// <see cref="clearCommand"/>
        /// </value>
        private IDbCommand ClearCommand
        {
            get
            {
                if (clearCommand == null || clearCommand.Connection != Connection)
                {
                    if (clearCommand != null)
                    {
                        // Close, but never fail
                        try
                        {
                            clearCommand.Dispose();
                        }
                        catch (Exception)
                        {
                        }
                    }

                    clearCommand = Connection.CreateCommand();
                    clearCommand.CommandText = "DELETE FROM Location; DELETE FROM Word; DELETE FROM Document";

                    // Preparing is often unneeded in modern systems, but it should not hurt, we're doing this once
                    // per connection only.
                    clearCommand.Prepare();
                }

                if (clearCommand.Transaction != transaction)
                {
                    clearCommand.Transaction = transaction;
                }

                return clearCommand;
            }
        }

        /// <summary>
        /// This member variable holds the loaded names and document indices loaded from the XML file that this
        /// instance is working against. The keys in this dictionary are the names, given by the user, of each document
        /// index, and the values are the <see cref="IndexInfo"/> instances, created in the <see cref="Load"/> method,
        /// which can be accessed by the user. This collection is accessed using the accessors in this class, and it is
        /// populated by the <see cref="Load"/> method.
        /// </summary>
        private IDictionary indices;

        /// <summary>
        /// This constructor creates a new <see cref="DBCollectionIndexSource"/> based on configuration parameters read
        /// from the namespace given. The parameters read should include the connection name.
        /// </summary>
        /// <param name="nameSpace">The filepath to use when loading.</param>
        /// <exception cref="ArgumentException">If <paramref name="nameSpace"/> is empty.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="nameSpace"/> is null.</exception>
        /// <exception cref="DBPersistenceConfigurationException">Any namespace or parameter missing.</exception>
        public DBCollectionIndexSource(string nameSpace)
        {
            if (nameSpace == null)
            {
                throw new ArgumentNullException("nameSpace");
            }
            if (nameSpace.Trim().Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "nameSpace");
            }

            indices = new Hashtable();

            try
            {
                ConfigManager cm = ConfigManager.GetInstance();
                parameters = new DBDocumentIndexParameters(
                    cm.GetValue(nameSpace, DBPersistenceParameters.CONNECTIONNAME_PROPERTY).Trim(),
                    long.Parse(cm.GetValue(nameSpace, DBDocumentIndexParameters.DOCUMENTID_PROPERTY)));
            }
            catch (Exception e)
            {
                throw new DBPersistenceConfigurationException("Configuration error.", e);
            }
        }

        /// <summary>
        /// Default constructor. A default namespace is used. No loading is performed.
        /// </summary>
        public DBCollectionIndexSource()
            : this(DBDocumentIndexSource.DefaultNamespace)
        {
        }

        /// <summary>
        /// This method adds the given <see cref="IndexInfo"/> into the indices dictionary, using the name given as the
        /// key.
        /// </summary>
        /// <param name="indexName">The name of the index to add.</param>
        /// <param name="indexInfo">The index info to add.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexInfo"/> or <paramref name="indexName"/> is
        /// null.</exception>
        /// <exception cref="ArgumentException">If the name in <paramref name="indexName"/> is empty, or is already in
        /// use.</exception>
        public void Add(string indexName, IndexInfo indexInfo)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty name string specified.", "indexName");
            }

            ValidateIndexInfo(indexInfo, "indexInfo");
            try
            {
                ((IDbDataParameter) AddDocument.Parameters[0]).Value =
                        (indexInfo.Parameters as DBDocumentIndexParameters).DocumentID;
                ((IDbDataParameter) AddDocument.Parameters[1]).Value = indexName;

                AddDocument.ExecuteNonQuery();
            }
            catch (Exception e)
            {
                throw new SaveIndexDataException(e, "Error adding.");
            }
            indices.Add(indexName, indexInfo);
        }

        /// <summary>
        /// This method removes the given index and all word locations belonging to it.
        /// </summary>
        /// <param name="indexName">The index to remove.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is an empty string.</exception>
        public void Remove(string indexName)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty name string specified.", "indexName");
            }

            IndexInfo info = (IndexInfo) indices[indexName];
            if (info != null)
            {
                try
                {
                    // Remove the document entry, and remove the locations. A cascading remove option in the schema
                    // ensures that the corresponding data is also deleted efficiently.

                    ((IDbDataParameter) RemoveDocument.Parameters[0]).Value =
                        (info.Parameters as DBDocumentIndexParameters).DocumentID;
                    ((IDbDataParameter) RemoveDocument.Parameters[1]).Value = indexName;

                    RemoveDocument.ExecuteNonQuery();
                }
                catch (Exception e)
                {
                    throw new SaveIndexDataException(e, "Error removing.");
                }

                indices.Remove(indexName);
            }
        }

        /// <summary>
        /// This method returns whether or not an index with the specified name is present.
        /// </summary>
        /// <param name="indexName">The index to check for.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is an empty string.</exception>
        /// <returns>True if it exists, false if not.</returns>
        public bool Contains(string indexName)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "indexName");
            }

            return indices.Contains(indexName);
        }

        /// <summary>
        /// This method clears all indices, including their content!
        /// </summary>
        /// <exception cref="SaveIndexDataException">Any SQL error.</exception>
        public void Clear()
        {
            try
            {
                ClearCommand.ExecuteNonQuery();
            }
            catch (Exception e)
            {
                throw new SaveIndexDataException(e, "Error clearing.");
            }
            indices.Clear();
        }

        /// <summary>
        /// This method just returns the enumerator for the indices collection,
        /// <see cref="IDictionary.GetEnumerator"/>. This method can be used by the user to enumerate through all the
        /// registered indices.
        /// </summary>
        /// <remarks>
        /// Please note that using an enumerator is always inherently thread-unsafe, if any
        /// editing is performed on the underlying structure. Lock on <c>SyncRoot</c> while consuming the enumerator,
        /// and avoid making any changes from the current thread, if needed.
        /// </remarks>
        /// <returns>
        /// An enumerator instance.
        /// </returns>
        public IDictionaryEnumerator GetEnumerator()
        {
            // We can't synchronize this any better than what's performed by the collection
            return indices.GetEnumerator();
        }

        /// <summary>
        /// This method loads all document index names and <see cref="IndexInfo"/> instances corresponding to those
        /// names based on the contents of the Document table.
        /// </summary>
        /// <exception cref="ArgumentNullException">If <paramref name="parameters"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="parameters"/> is not a
        /// <see cref="DBPersistenceParameters"/> instance.
        /// </exception>
        /// <exception cref="LoadIndexDataException">Any error while loading the indices.</exception>
        /// <param name="parameters">The parameters specifying the DB to load from.</param>
        public void Load(IParameters parameters)
        {
            if (parameters == null)
            {
                throw new ArgumentNullException("parameters");
            }
            if (!(parameters is DBPersistenceParameters))
            {
                throw new ArgumentException("Specified parameters not of DBPersistenceParameters type.",
                    "parameters");
            }

            this.parameters = (DBPersistenceParameters) parameters;

            // Our Dispose will just close the current connection, but the object will remain usable.
            Dispose();
            // A new connection will be created, based on the current parameters.
            try
            {
                indices.Clear();
                using (IDataReader reader = LoadDocuments.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        // Unboxing needed, if the type is Int32 according to the reader, this will fail if we just
                        // used a (long) cast
                        long document_id = Convert.ToInt64(reader[0]);
                        // Trim needed especially if CHAR is used (not VARCHAR)
                        string document_name = ((string) reader[1]).Trim();

                        indices[document_name] = new IndexInfo(typeof(DBDocumentIndexSource).Assembly.FullName,
                            typeof(DBDocumentIndexSource).FullName,
                            new DBDocumentIndexParameters(this.parameters.ConnectionName, document_id));
                    }
                }
            }
            catch (Exception e)
            {
                // Avoid simulating that a Load actually took place, when it did not.
                // This will create an undefined state, but it can be solved by a following, valid, Load.
                parameters = null;
                indices.Clear();
                Dispose();

                throw new LoadIndexDataException(parameters, e);
            }
            finally
            {
                // Try to commit the current transaction, by using Save. If we didn't do this, the current transaction
                // would always include the huge Load operation, which could easily block other transactions, depending
                // on the concurrency algorithm and lock granularity.
                //
                // However, if this fails, it's not a catastrophic failure in itself. We're just trying to be kind.
                try
                {
                    Save();
                }
                catch (Exception)
                {
                }
            }
        }

        /// <summary>
        /// This method commits any changes made in the current transaction. Note that updates in other indexes may
        /// fail if they would affect data that's already been affected by this transaction.
        /// </summary>
        /// <exception cref="SaveIndexDataException">Any error while saving.</exception>
        public void Save()
        {
            try
            {
                // Nope, saving something that hasn't started won't do much good
                if (connection != null)
                {
                    if (transaction != null)
                    {
                        transaction.Commit();
                        transaction.Dispose();
                    }
                    transaction = Connection.BeginTransaction();
                }
            }
            catch (Exception e)
            {
                throw new SaveIndexDataException(e, "Error during transaction commit.");
            }
        }

        /// <summary>
        /// Gets/sets the <see cref="IndexInfo"/> of the index with the specified name.
        /// </summary>
        /// <param name="indexName">The index to change/read the info for..</param>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is empty.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null (only setter).</exception>
        /// <exception cref="ArgumentException">If the passed-in <see cref="IndexInfo"/> doesn't seem valid, for this
        /// type of collection.
        /// </exception>
        /// <exception cref="SaveIndexDataException">Any SQL error while updating the database representation.
        /// </exception>
        /// <value>
        /// An <see cref="IndexInfo"/>, or null in the getter if it's not defined for that name.
        /// </value>
        public IndexInfo this[string indexName]
        {
            get
            {
                if (indexName == null)
                {
                    throw new ArgumentNullException("indexName");
                }
                indexName = indexName.Trim();
                if (indexName.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "indexName");
                }

                return (IndexInfo) indices[indexName];
            }
            set
            {
                if (indexName == null)
                {
                    throw new ArgumentNullException("indexName");
                }

                indexName = indexName.Trim();
                if (indexName.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "indexName");
                }

                ValidateIndexInfo(value, "value");

                // Remove existing if needed
                if (indices.Contains(indexName))
                {
                    Remove(indexName);
                }

                Add(indexName, value);
                indices[indexName] = value;
            }
        }

        /// <summary>
        /// Validates the specified index info.
        /// </summary>
        /// <param name="value">The value to validate.</param>
        /// <param name="paramName">The parameter name to use in exceptions.</param>
        /// <exception cref="ArgumentNullException">Null values.</exception>
        /// <exception cref="ArgumentException">Any parameter error, except for null.</exception>
        private void ValidateIndexInfo(IndexInfo value, string paramName)
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName);
            }

            Type type;
            try
            {
                type = Assembly.Load(value.AssemblyName).GetType(value.ClassName);
            }
            catch (Exception e)
            {
                throw new ArgumentException("Could not determine identity of referenced type.", paramName, e);
            }

            if (!typeof(DBDocumentIndexSource).Equals(type))
            {
                throw new ArgumentException("Incorrect type referenced in index info.", paramName);
            }

            DBDocumentIndexParameters parameters = value.Parameters as DBDocumentIndexParameters;
            if (parameters == null)
            {
                throw new ArgumentException("Incorrect IParameters implementation used.", paramName);
            }

            if (parameters.ConnectionName != this.parameters.ConnectionName)
            {
                throw new ArgumentException("This IndexInfo uses a different connection name.", paramName);
            }
        }

        /// <summary>
        /// Disposes the current connection. The active transaction is rolled back. However, except for this, the
        /// database connection will be resurrected when loading again. Note that any changes to the local
        /// representation are NOT rolled back, so disposing an object with active changes will create a
        /// synchronization mismatch.
        /// </summary>
        /// <remarks>
        /// We do only encapsulate other managed resources, so a separate finalizer wouldn't do any good.
        /// </remarks>
        public void Dispose()
        {
            try
            {
                if (connection != null)
                {
                    if (transaction != null)
                    {
                        transaction.Rollback();
                        transaction.Dispose();
                    }
                    connection.Dispose();
                    connection = null;
                    transaction = null;
                }
            }
            catch (Exception)
            {
            }
        }
    }
}
