/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using System;
using System.Collections;
using System.IO;
using System.Text;
using System.Reflection;
using System.Xml;
using TopCoder.Document.Index.Persistence;

namespace TopCoder.Document.Index.Persistence.Xml
{
    /// <summary>
    /// This class implements the <see cref="ICollectionIndexSource"/> to provide the sources in a XML document.
    /// </summary>
    /// <remarks>
    /// This implementation saves and parses files according to the "Documents.xsd" file in the docs/ directory. This
    /// class provides functionality for getting information about a particular document, as well as saving back out
    /// all documents in the collection. The indices member variable in this class is mutable, but access to it is
    /// synchronized, leading to a thread safe implementation.
    /// </remarks>
    public class XmlCollectionIndexSource : ICollectionIndexSource
    {
        /// <summary>
        /// The Uri of the namespace for all XML files.
        /// </summary>
        private const string XML_NAMESPACE = "http://www.topcoder.com";

        /// <summary>
        /// The root of the main XML file for a document index.
        /// </summary>
        private const string DOCUMENTS_ELEMENT = "Documents";

        /// <summary>
        /// Each document is represented by a specific element, listing name and main XML file.
        /// </summary>
        private const string DOCUMENT_ELEMENT = "Document";

        /// <summary>
        /// The name element for a document.
        /// </summary>
        private const string NAME_ELEMENT = "Name";

        /// <summary>
        /// The content of this element is the file path to a specific file, specified as a path relative to that
        /// of the main file.
        /// </summary>
        private const string FILEPATH_ELEMENT = "FilePath";

        /// <summary>
        /// This member variable holds the path to the file where XML is loaded from and saved to. This member variable
        /// can be set in the constructor, and it can be set and retrieved using the <see cref="FilePath"/> property.
        /// It's also updated by <see cref="Load"/> and used by <see cref="Save"/>.
        /// </summary>
        private string filePath = null;

        /// <summary>
        /// This property provides access to the internally stored file path, used to set the location where the XML
        /// will be saved.
        /// </summary>
        /// <exception cref="ArgumentException">If <paramref name="value"/> is empty.</exception>
        /// <exception cref="ArgumentException">If <paramref name="value"/> is an invalid path, for example pointing
        /// to a directory.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null.</exception>
        /// <value>
        /// A non-null, non-empty path.
        /// </value>
        public string FilePath
        {
            get
            {
                return filePath;
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                value = value.Trim();
                if (value.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", value);
                }

                try
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(value);
                    if (dirInfo.Exists)
                    {
                        throw new ArgumentException("Path points to a directory.");
                    }
                }
                catch (ArgumentException)
                {
                    throw;
                }
                catch (Exception e)
                {
                    throw new ArgumentException("Path invalid or access denied.", "value", e);
                }

                // Don't possible do anything with the current path while loading/saving
                lock (SyncRoot)
                {
                    filePath = value;
                }
            }
        }

        /// <summary>
        /// This member variable holds the loaded names and document indices loaded from the XML file that this
        /// instance is working against. The keys in this dictionary are the names, given by the user, of each document
        /// index, and the values are the <see cref="IndexInfo"/> instances, created in the <see cref="Load"/> method,
        /// which can be accessed by the user. This collection is accessed using the accessors in this class, and it is
        /// populated by the <see cref="Load"/> method.
        /// </summary>
        private IDictionary indices;

        /// <summary>
        /// This property provides the object to use when synchronizing access to the indices collection.
        /// </summary>
        /// <value>
        /// A reference to the sync root on the internal collection.
        /// </value>
        public object SyncRoot
        {
            get
            {
                return indices.SyncRoot;
            }
        }

        /// <summary>
        /// This constructor initializes the class using the <paramref name="filePath"/> given. That property is the
        /// path where the XML data is saved to. No loading is performed in the constructor.
        /// </summary>
        /// <param name="filePath">The filepath to use when loading.</param>
        /// <exception cref="ArgumentException">If <paramref name="filePath"/> is empty.</exception>
        /// <exception cref="ArgumentException">If <paramref name="filePath"/> is an invalid path, for example pointing
        /// to a directory.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="filePath"/> is null.</exception>
        public XmlCollectionIndexSource(string filePath)
            : this()
        {
            FilePath = filePath;
        }

        /// <summary>
        /// Default constructor, intended to be used with <see cref="Load"/> to load a specific file afterwards.
        /// </summary>
        public XmlCollectionIndexSource()
        {
            indices = Hashtable.Synchronized(new Hashtable());
        }

        /// <summary>
        /// This method adds the given <see cref="IndexInfo"/> into the indices dictionary, using the name given as the
        /// key.
        /// </summary>
        /// <param name="indexName">The name of the index to add.</param>
        /// <param name="indexInfo">The index info to add.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexInfo"/> or <paramref name="indexName"/> is
        /// null.</exception>
        /// <exception cref="ArgumentException">If the name in <paramref name="indexName"/> is empty, or is already in
        /// use.</exception>
        public void Add(string indexName, IndexInfo indexInfo)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty name string specified.", "indexName");
            }
            ValidateIndexInfo(indexInfo, "indexInfo");

            indices.Add(indexName, indexInfo);
        }

        /// <summary>
        /// This method removes the given index from the indices collection.
        /// </summary>
        /// <param name="indexName">The index to remove.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is an empty string.</exception>
        public void Remove(string indexName)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty name string specified.", "indexName");
            }

            // The collection is synchronizing itself, common to our SyncRoot
            indices.Remove(indexName);
        }

        /// <summary>
        /// This method returns whether or not an index with the specified name is present.
        /// </summary>
        /// <param name="indexName">The index to check for.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is an empty string.</exception>
        /// <returns>True if it exists, false if not.</returns>
        public bool Contains(string indexName)
        {
            if (indexName == null)
            {
                throw new ArgumentNullException("indexName");
            }
            indexName = indexName.Trim();
            if (indexName.Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "indexName");
            }

            return indices.Contains(indexName);
        }

        /// <summary>
        /// This method clears all indices.
        /// </summary>
        public void Clear()
        {
            // Synchronization is handled by the internal object
            indices.Clear();
        }

        /// <summary>
        /// This method just returns the enumerator for the indices collection,
        /// <see cref="IDictionary.GetEnumerator"/>. This method can be used by the user to enumerate through all the
        /// registered indices.
        /// </summary>
        /// <remarks>
        /// Please note that using an enumerator is always inherently thread-unsafe, if any
        /// editing is performed on the underlying structure. Lock on <c>SyncRoot</c> while consuming the enumerator,
        /// and avoid making any changes from the current thread, if needed.
        /// </remarks>
        /// <returns>
        /// An enumerator instance.
        /// </returns>
        public IDictionaryEnumerator GetEnumerator()
        {
            // We can't synchronize this any better than what's performed by the collection
            return indices.GetEnumerator();
        }

        /// <summary>
        /// Reads until an element start, or a file end, is found.
        /// </summary>
        /// <param name="reader">The reader to read from.</param>
        private static void ReadToStartElement(XmlReader reader)
        {
            while (reader.NodeType != XmlNodeType.Element && reader.Read())
            {
            }
        }

        /// <summary>
        /// This method loads all document index names and <see cref="IndexInfo"/> instances corresponding to those
        /// names, from the XML file path that is specified in the <see cref="SimpleIndexParameters"/> instance
        /// provided. The XML in the file path given should follow the Documents schema outlined in the docs folder.
        /// </summary>
        /// <remarks>
        /// Please note that the state will be undefined if the load operation fails.
        /// </remarks>
        /// <exception cref="ArgumentNullException">If <paramref name="parameters"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="parameters"/> is not a
        /// <see cref="SimpleIndexParameters"/> instance, or the path is invalid, or the file doesn't exist.
        /// </exception>
        /// <exception cref="LoadIndexDataException">Any error while parsing XML.</exception>
        /// <param name="parameters">The parameters specifying the path to load from.</param>
        public void Load(IParameters parameters)
        {
            if (parameters == null)
            {
                throw new ArgumentNullException("parameters");
            }

            SimpleIndexParameters pathParams = parameters as SimpleIndexParameters;
            if (pathParams == null)
            {
                throw new ArgumentException("Specified parameters not of SimpleIndexParameters type.", "parameters");
            }

            lock (SyncRoot)
            {
                string relativePath = null;

                try
                {
                    if (!File.Exists(pathParams.FileName))
                    {
                        throw new ArgumentException("The specified file doesn't exist.", "parameters");
                    }
                    relativePath = Path.GetDirectoryName(pathParams.FileName);
                }
                catch (ArgumentException)
                {
                    throw;
                }
                catch (Exception e)
                {
                    throw new ArgumentException("Error accessing file.", "parameters", e);
                }

                FilePath = pathParams.FileName;
                Clear();

                XmlReader reader = null;
                try
                {
                    reader = new XmlTextReader(filePath);

                    bool readRoot = false;
                    while (reader.Read())
                    {
                        if (reader.NodeType == XmlNodeType.Element)
                        {
                            // Document
                            reader.ReadStartElement(DOCUMENTS_ELEMENT, XML_NAMESPACE);

                            if (readRoot)
                            {
                                throw new XmlException(String.Format("Double root elements in {0}.", filePath));
                            }
                            readRoot = true;


                            while (reader.NodeType != XmlNodeType.EndElement)
                            {
                                if (reader.NodeType == XmlNodeType.Element)
                                {
                                    reader.ReadStartElement(DOCUMENT_ELEMENT, XML_NAMESPACE);
                                    ReadToStartElement(reader);
                                    string name = reader.ReadElementString(NAME_ELEMENT, XML_NAMESPACE).Trim();
                                    ReadToStartElement(reader);
                                    string file = reader.ReadElementString(FILEPATH_ELEMENT, XML_NAMESPACE).Trim();

                                    // Add it, fill all relevant data for instantiation through reflection.
                                    indices[name] =
                                        new IndexInfo(typeof(XmlDocumentIndexSource).Assembly.FullName,
                                        typeof(XmlDocumentIndexSource).FullName,
                                        new SimpleIndexParameters(file));

                                    // Read until </Document>
                                    while (reader.NodeType != XmlNodeType.EndElement && reader.Read())
                                    {
                                    }
                                }

                                // Advance
                                reader.Read();
                            }
                        }
                    }
                    if (!readRoot)
                    {
                        throw new XmlException(String.Format("No root element found in {0}.", filePath));
                    }
                }
                catch (Exception e)
                {
                    // Avoid any user saving the now undefined data.
                    filePath = null;
                    Clear();
                    throw new LoadIndexDataException(parameters, e, "Error loading from XML.");
                }
                finally
                {
                    if (reader != null)
                    {
                        reader.Close();
                    }
                }
            }
        }

        /// <summary>
        /// This method saves out the internally stored <see cref="IDictionary"/> as a XML file, using the Documents
        /// schema.
        /// </summary>
        /// <exception cref="SaveIndexDataException">Any error while saving.</exception>
        /// <exception cref="InvalidOperationException">If saving is called in a state where no file name is present.
        /// </exception>
        public void Save()
        {
            if (FilePath == null)
            {
                throw new InvalidOperationException("Saving impossible if no file path has been specified.");
            }

            try
            {
                lock (SyncRoot)
                {
                    XmlWriter mainWriter = null;
                    try
                    {
                        mainWriter = new XmlTextWriter(filePath, Encoding.UTF8);
                        // Root element
                        mainWriter.WriteStartElement(DOCUMENTS_ELEMENT, XML_NAMESPACE);

                        // One entry for every single element
                        foreach (DictionaryEntry entry in indices)
                        {
                            mainWriter.WriteStartElement(DOCUMENT_ELEMENT, XML_NAMESPACE);
                            mainWriter.WriteElementString(NAME_ELEMENT, entry.Key.ToString());

                            mainWriter.WriteElementString(FILEPATH_ELEMENT,
                                ((entry.Value as IndexInfo).Parameters as SimpleIndexParameters).FileName);
                            mainWriter.WriteEndElement();
                        }

                        mainWriter.WriteEndElement();
                    }
                    // We can't put the "catch" here, since closing, in itself, may throw an exception
                    finally
                    {
                        if (mainWriter != null)
                        {
                            mainWriter.Close();
                        }
                    }
                }
            }
            catch (Exception e)
            {
                throw new SaveIndexDataException(e, "Error saving.");
            }

        }

        /// <summary>
        /// Gets/sets the <see cref="IndexInfo"/> of the index with the specified name.
        /// </summary>
        /// <param name="indexName">The index to change/read the info for..</param>
        /// <exception cref="ArgumentException">If <paramref name="indexName"/> is empty.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="indexName"/> is null.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null (only setter).</exception>
        /// <exception cref="ArgumentException">If the passed-in <see cref="IndexInfo"/> doesn't seem valid, for this
        /// type of collection.
        /// </exception>
        /// <value>
        /// An <see cref="IndexInfo"/>, or null in the getter if it's not defined for that name.
        /// </value>
        public IndexInfo this[string indexName]
        {
            get
            {
                if (indexName == null)
                {
                    throw new ArgumentNullException("indexName");
                }
                indexName = indexName.Trim();
                if (indexName.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "indexName");
                }

                return (IndexInfo) indices[indexName];
            }
            set
            {
                if (indexName == null)
                {
                    throw new ArgumentNullException("indexName");
                }

                indexName = indexName.Trim();
                if (indexName.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "indexName");
                }

                ValidateIndexInfo(value, "value");
                indices[indexName] = value;
            }
        }

        /// <summary>
        /// Validates the specified index info.
        /// </summary>
        /// <param name="value">The value to validate.</param>
        /// <param name="paramName">The parameter name to use in exceptions.</param>
        /// <exception cref="ArgumentNullException">Null values.</exception>
        /// <exception cref="ArgumentException">Any parameter error, except for null.</exception>
        private static void ValidateIndexInfo(IndexInfo value, string paramName)
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName);
            }

            Type type;
            try
            {
                type = Assembly.Load(value.AssemblyName).GetType(value.ClassName);
            }
            catch (Exception e)
            {
                throw new ArgumentException("Could not determine identity of referenced type.", paramName, e);
            }

            if (!typeof(XmlDocumentIndexSource).Equals(type))
            {
                throw new ArgumentException("Incorrect type referenced in index info.", paramName);
            }
            if (!(value.Parameters is SimpleIndexParameters))
            {
                throw new ArgumentException("Incorrect IParameters implementation used.", paramName);
            }
        }
    }
}
