/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER
 * Version: 1.0
 */

using NUnit.Framework;
using System;
using System.Collections;
using System.Globalization;
using TopCoder.Document.Index.WordStreams;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class tests <see cref="DBDocumentIndexSource"/>.
    /// </summary>
    /// <remarks>
    /// This class "implements" <see cref="IParameters"/> to supply itself as an (incorrect) parameter object.
    /// </remarks>
    [TestFixture]
    public class TestDBDocumentIndexSource : DBTests, IParameters
    {
        /// <summary>
        /// A test object used in several tests, created based on a populated index.
        /// </summary>
        private DBDocumentIndexSource testObject1;

        /// <summary>
        /// A test object crated with an empty index.
        /// </summary>
        private DBDocumentIndexSource testObject2;

        /// <summary>
        /// A reference to the type used, for testing.
        /// </summary>
        private static readonly Type type = typeof(DBDocumentIndexSource);

        /// <summary>
        /// Sets up the testing fixture.
        /// </summary>
        [SetUp]
        public override void Setup()
        {
            base.Setup();

            testObject1 = new DBDocumentIndexSource();
            testObject1.Load(new DBDocumentIndexParameters("Test", 1));

            testObject2 = new DBDocumentIndexSource();
        }

        /// <summary>
        /// Counts the elements reported by an enumerator.
        /// </summary>
        /// <param name="enumerator">The enumerator.</param>
        /// <returns>The number of elements.</returns>
        private static int CountEnumerator(IEnumerator enumerator)
        {
            int count = 0;
            while (enumerator.MoveNext())
            {
                count++;
            }

            return count;
        }

        /// <summary>
        /// Tests the <see cref="DBDocumentIndexSource.IndexInfo"/> implementation. This property is a semi-constant
        /// getter-only. It's used to be able to reinstantiate an index, through reflection.
        /// </summary>
        [Test]
        public void TestIndexInfo()
        {
            IndexInfo info = testObject1.IndexInfo;

            Assert.AreEqual(type.Assembly.FullName, info.AssemblyName, "Assembly name mismatch.");
            Assert.AreEqual(type.FullName, info.ClassName, "Class name mismatch.");

            DBDocumentIndexParameters parameters = info.Parameters as DBDocumentIndexParameters;

            Assert.AreEqual(1, parameters.DocumentID, "The document ID doesn't match.");
            Assert.AreEqual("Test", parameters.ConnectionName, "The connection name doesn't match.");
        }

        /// <summary>
        /// Asks for a document index based on the <see cref="IndexInfo"/> returned.
        /// </summary>
        [Test]
        public void TestIndexInfoForReal()
        {
            DocumentIndex docIndex = new IndexFactory().GetIndex(testObject1.IndexInfo);
            Assert.IsFalse(docIndex.Contains("ENTRU"), "ENTRU should not be present.");
            Assert.IsTrue(docIndex.Contains("entr"), "entr should be present.");
            Assert.IsTrue(docIndex.Contains("entre"), "entre should be present (equal to entr).");
        }

        /// <summary>
        /// Tests the <see cref="DBDocumentIndexSource.Culture"/> implementation. This property has both a getter and
        /// setter. The default value is verified, and then the value is changed.
        /// </summary>
        [Test]
        public void TestCulture()
        {
            Assert.AreEqual(new CultureInfo("en-us"), testObject2.Culture, "Default value not returned.");
            testObject2.Culture = new CultureInfo("fr-FR");

            Assert.AreEqual(new CultureInfo("fr-FR"), testObject2.Culture, "Changed value not returned.");
            Assert.AreEqual(new CultureInfo("fr-FR"), testObject1.Culture, "Database value not returned.");
        }

        /// <summary>
        /// Tests setting a null culture. This is, naturally, not allowed. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestCultureNull()
        {
            testObject1.Culture = null;
        }

        /// <summary>
        /// Tests changing culture for a non-empty index. This is not allowed.
        /// </summary>
        [Test]
        [ExpectedException(typeof(InvalidOperationException))]
        public void TestCultureNonEmpty()
        {
            testObject1.Culture = CultureInfo.InvariantCulture;
        }

        /// <summary>
        /// Tests the <see cref="DBDocumentIndexSource.CompareOptions"/> implementation. This property has both a
        /// getter and setter. The default value is verified, and then the value is changed.
        /// </summary>
        [Test]
        public void TestCompareOptions()
        {
            Assert.AreEqual(CompareOptions.None, testObject2.CompareOptions, "Default value not returned.");
            testObject2.CompareOptions = CompareOptions.IgnoreNonSpace;

            Assert.AreEqual(CompareOptions.IgnoreNonSpace, testObject2.CompareOptions, "Changed value not returned.");
            Assert.AreEqual(CompareOptions.IgnoreNonSpace, testObject1.CompareOptions, "Database value not returned.");
        }

        /// <summary>
        /// Tests changing compare options for a non-empty index. This is not allowed.
        /// </summary>
        [Test]
        [ExpectedException(typeof(InvalidOperationException))]
        public void TestCompareOptionsNonEmpty()
        {
            testObject1.CompareOptions = CompareOptions.IgnoreCase;
        }

        /// <summary>
        /// Tests constructing with a null namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestNameSpaceConstructorNull()
        {
            new DBDocumentIndexSource(null);
        }

        /// <summary>
        /// Tests constructing with an empty namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestNameSpaceConstructorEmpty()
        {
            new DBDocumentIndexSource(" ");
        }

        /// <summary>
        /// Tests constructing with a non-existing namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(DBPersistenceConfigurationException))]
        public void TestNameSpaceConstructorInvalid()
        {
            new DBDocumentIndexSource("Some.Weird.Namespace");
        }

        /// <summary>
        /// Tests constructing with an existing namespace, but not the required parameters.
        /// </summary>
        [Test]
        [ExpectedException(typeof(DBPersistenceConfigurationException))]
        public void TestNameSpaceConstructorMissingParam()
        {
            new DBDocumentIndexSource("TopCoder.Document.Index.Persistence.NoID");
        }

        /// <summary>
        /// Tests the namespace constructor, verifying that the document ID and connection name are correctly read.
        /// </summary>
        [Test]
        public void TestNameSpaceConstructor()
        {
            DBDocumentIndexSource source = new DBDocumentIndexSource("TopCoder.Document.Index.Persistence.DB.2");
            DBDocumentIndexParameters info = source.IndexInfo.Parameters as DBDocumentIndexParameters;

            Assert.AreEqual(3, info.DocumentID, "The document ID doesn't match.");
            Assert.AreEqual("AnotherConnection", info.ConnectionName, "The connection name doesn't match.");
        }

        /// <summary>
        /// Tests adding words, both already existing ones (with different representations) and new ones.
        /// </summary>
        [Test]
        public void TestAdd()
        {
            WordInfo word = new WordInfo("entr", 999);
            testObject1.Add(word);

            // Diacritic-insensitive comparison
            IList list = testObject1["entre"];
            Assert.AreEqual(3, list.Count, "List count not as expected. 1");
            Assert.AreEqual(1, list[0], "First position not as expected. 1");
            Assert.AreEqual(3, list[1], "Second position not as expected. 1");
            Assert.AreEqual(999, list[2], "Third position not as expected. 1");

            // We need to commit the actions on object 1 to allow writes to object 2
            testObject1.Save();

            testObject2.Add(word);
            // Diacritic-SENSITIVE comparison
            Assert.IsNull(testObject2["entre"], "Entre shouldn't match entr. 2");

            list = testObject2["entr"];
            Assert.AreEqual(1, list.Count, "Exactly one element expected. 2");
            Assert.AreEqual(999, list[0], "Position not as expected. 2");
        }

        /// <summary>
        /// Tests adding a null word. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestAddNull()
        {
            testObject1.Add(null);
        }

        /// <summary>
        /// Tests adding an empty word. Exception expected. The definition of an empty string is different in the
        /// document indexer and this component.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddEmpty()
        {
            testObject1.Add(new WordInfo(" ", 0));
        }

        /// <summary>
        /// Tests removing a word, both one which really exists and one which does not.
        /// </summary>
        [Test]
        public void TestRemove()
        {
            testObject1.Remove("entre");
            Assert.IsNull(testObject1["entr"], "Entr should now be a null list.");

            Assert.AreEqual(1, CountEnumerator(testObject1.GetEnumerator()), "Word count not as expected. 1");

            // The count should remain zero, but no exception shold be thrown
            testObject2.Remove("entre");
            Assert.IsNull(testObject2["entre"], "Entre should now be a null list.");
            Assert.AreEqual(0, CountEnumerator(testObject2.GetEnumerator()), "Word count not as expected. 2");
        }

        /// <summary>
        /// Tests removing a null word. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestRemoveNull()
        {
            testObject1.Remove(null);
        }

        /// <summary>
        /// Tests removing an empty word. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestRemoveEmpty()
        {
            testObject1.Remove(" ");
        }

        /// <summary>
        /// A simple test of <see cref="DBDocumentIndexSource.Contains"/>.
        /// </summary>
        [Test]
        public void TestContains()
        {
            Assert.IsTrue(testObject1.Contains("entre"), "Entr should exist.");
            Assert.IsFalse(testObject2.Contains("entre"), "Entre should not exist.");
        }

        /// <summary>
        /// Another test of <see cref="DBDocumentIndexSource.Contains"/>. An empty list means that the method should
        /// return false.
        /// </summary>
        [Test]
        public void TestContainsEmptyList()
        {
            testObject1["entr"] = new int[0];
            Assert.IsFalse(testObject1.Contains("entre"), "Entr should not exist, when the list is empty.");
        }

        /// <summary>
        /// Tests checking for a null word. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestContainsNull()
        {
            testObject1.Contains(null);
        }

        /// <summary>
        /// Tests checking for an empty word. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestContainsEmpty()
        {
            testObject1.Contains(" ");
        }

        /// <summary>
        /// Tests clearing the store.
        /// </summary>
        [Test]
        public void TestClear()
        {
            testObject1.Clear();
            Assert.IsNull(testObject1["entr"], "Known word should now be null.");
            Assert.AreEqual(0, CountEnumerator(testObject1.GetEnumerator()), "Count should be zero after clearing.");
        }

        /// <summary>
        /// Tests <see cref="DBDocumentIndexSource.GetEnumerator"/>, by iterating in order.
        /// </summary>
        [Test]
        public void TestGetEnumerator()
        {
            IDictionaryEnumerator enumerator = testObject1.GetEnumerator();
            for (int i = 0; i < 2; i++)
            {
                enumerator.MoveNext();
                Assert.AreEqual("entre", enumerator.Key, "The first key doesn't match.");
                Assert.AreEqual(2, ((IList) enumerator.Value).Count, "Wrong count of first word.");

                enumerator.MoveNext();
                Assert.AreEqual("rouge", enumerator.Key, "The second key doesn't match.");
                Assert.AreEqual(1, ((IList) enumerator.Value).Count, "Wrong count of second word.");

                Assert.IsFalse(enumerator.MoveNext(), "Moving next should fail after end.");
                enumerator.Reset();
            }
        }

        /// <summary>
        /// Tests loading new data, after first adding a word.
        /// </summary>
        [Test]
        public void TestLoad()
        {
            testObject1.Dispose();

            WordInfo word = new WordInfo("entr", 999);
            testObject2.Add(word);
            testObject2.Load(new DBDocumentIndexParameters("Test", 1));

            testObject1 = testObject2;

            // Will verify that testObject1 now matches testObject2
            TestGetEnumerator();

            // Now, reload the original document
            testObject2.Load(new DBDocumentIndexParameters("Test", 2));
            Assert.AreEqual(0, CountEnumerator(testObject2.GetEnumerator()),
                "Word added before loading should be gone.");
        }

        /// <summary>
        /// Tests loading with a null reference. Exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestLoadNull()
        {
            testObject1.Load(null);
        }

        /// <summary>
        /// Tests loading with a parameter object of incorrect type.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestLoadWrongType()
        {
            // this implements IParameters
            testObject1.Load(this);
        }

        /// <summary>
        /// It should be possible to save the default object.
        /// </summary>
        [Test]
        public void TestSaveDefault()
        {
            testObject2.Save();
        }

        /// <summary>
        /// Add a word, commit, dispose, load in a new object and verify the results.
        /// </summary>
        [Test]
        public void TestSave()
        {
            WordInfo word = new WordInfo("entr", 999);
            testObject1.Add(word);

            // Save for testing, Dispose for resource management
            testObject1.Save();
            testObject1.Dispose();

            testObject1 = new DBDocumentIndexSource();
            testObject1.Load(new DBDocumentIndexParameters("Test", 1));

            // Diacritic-insensitive comparison
            object[] list = new ArrayList(testObject1["entre"]).ToArray();
            Assert.AreEqual(3, list.Length, "List count not as expected. 1");

            // Make us independent of the order, which the DBMS may rearrange
            Array.Sort(list);

            Assert.AreEqual(1, list[0], "First position not as expected. 1");
            Assert.AreEqual(3, list[1], "Second position not as expected. 1");
            Assert.AreEqual(999, list[2], "Third position not as expected. 1");
        }

        /// <summary>
        /// Add a word, dispose, load in a new object and verify that nothing has changed. (Dispose will roll-back.)
        /// </summary>
        [Test]
        public void TestDispose()
        {
            WordInfo word = new WordInfo("entr", 999);
            testObject1.Add(word);
            testObject1.Dispose();

            testObject1 = new DBDocumentIndexSource();
            testObject1.Load(new DBDocumentIndexParameters("Test", 1));

            // Diacritic-insensitive comparison
            IList list = testObject1["entre"];
            Assert.AreEqual(2, list.Count, "List count not as expected. 1");
            Assert.AreEqual(1, list[0], "First position not as expected. 1");
            Assert.AreEqual(3, list[1], "Second position not as expected. 1");
        }

        /// <summary>
        /// Some tests for comparison, all with valid string parameters. Non-string arguments, resulting in exceptions,
        /// are tested in separate cases.
        /// </summary>
        [Test]
        public void TestCompare()
        {
            Assert.AreEqual(0, testObject1.Compare("entre", "entre"), "Identical strings are identical. 1");
            Assert.AreEqual(0, testObject1.Compare("entr", "entre"), "Identity depends on comparison options. 1");
            Assert.IsTrue(testObject1.Compare("B", "A") > 0, "B > A 1");
            Assert.IsTrue(testObject1.Compare("A", "B") < 0, "A < B 1");

            Assert.AreEqual(0, testObject2.Compare("entre", "entre"), "Identical strings are identical. 2");
            Assert.IsTrue(testObject2.Compare("entr", "entre") > 0, "Identity depends on comparison options. 2");
            Assert.IsTrue(testObject2.Compare("B", "A") > 0, "B > A 2");
            Assert.IsTrue(testObject2.Compare("A", "B") < 0, "A < B 2");
        }

        /// <summary>
        /// Parameters being null, compared against strings. Null is less than everything, except null.
        /// </summary>
        [Test]
        public void TestCompareNull()
        {
            Assert.IsTrue(testObject2.Compare(null, "") < 0, "Null should compare as less than empty string. 1");
            Assert.IsTrue(testObject2.Compare("", null) > 0, "Null should compare as less than empty string. 2");

            Assert.IsTrue(testObject2.Compare(null, null) == 0, "Null is equal to null.");
        }

        /// <summary>
        /// A null value compared against a non-string raises an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCompareNullAgainstNonString1()
        {
            testObject2.Compare(null, this);
        }

        /// <summary>
        /// A null value compared against a non-string raises an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestCompareNullAgainstNonString2()
        {
            testObject2.Compare(this, null);
        }

        /// <summary>
        /// The getter part of the indexer is quite simple. One of the more important elements is to check that the
        /// returned value is only a clone, and nothing else.
        /// </summary>
        [Test]
        public void TestGetIndexer()
        {
            for (int i = 0; i < 2; i++)
            {
                IList list = testObject1["entre"];
                Assert.AreEqual(2, list.Count, "Two elements should be present.");
                Assert.AreEqual(1, list[0], "First position not as expected.");
                Assert.AreEqual(3, list[1], "Second position not as expected.");

                // Change the first entry, loop again, verifying that the stored data do not change
                list[0] = 99999;
            }

            Assert.IsNull(testObject1["I DON'T EXIST"],
                "Null should be returned for those words that are not present.");
        }

        /// <summary>
        /// Tests getting the list for a null word.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestGetIndexerNull()
        {
            IList list = testObject1[null];
        }

        /// <summary>
        /// Tests getting the list for an empty word.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestGetIndexerEmpty()
        {
            IList list = testObject1[" "];
        }

        /// <summary>
        /// Tests setting lists with the indexer, including that the type is always changed to <see cref="long"/>.
        /// </summary>
        [Test]
        public void TestSetIndexer()
        {
            testObject1["someword"] = new int[] { 2, 3 };
            IList list = testObject1["someword"];

            Assert.AreEqual(2, list[0], "The first value not as expected.");
            Assert.AreEqual(3, list[1], "The second value not as expected.");

            for (int i = 0; i < 2; i++)
            {
                Assert.AreEqual(typeof(Int64), list[i].GetType(), "Element type not as expected.");
            }
        }

        /// <summary>
        /// Tests setting the list for a null word. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullWord()
        {
            testObject1[null] = new long[0];
        }

        /// <summary>
        /// Tests setting the list for an empty word. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestSetIndexerEmptyWord()
        {
            testObject1[" "] = new long[0];
        }

        /// <summary>
        /// Tests setting a null list. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullList()
        {
            testObject1["someword"] = null;
        }

        /// <summary>
        /// Closes active connections. This was not specified by the design, but it is dearly needed if the connection
        /// limit on the Informix server is low.
        /// </summary>
        [TearDown]
        public override void Teardown()
        {
            testObject1.Dispose();
            testObject2.Dispose();
            base.Teardown();
        }
    }
}
