/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER
 * Version: 1.0
 */

using NUnit.Framework;
using System;
using System.Collections;
using System.Globalization;
using TopCoder.Document.Index.WordStreams;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class tests <see cref="DBPersistenceParameters"/>, it's also a base class to
    /// <see cref="TestDBDocumentIndexParameters"/>.
    /// </summary>
    [TestFixture]
    public class TestDBPersistenceParameters
    {
        /// <summary>
        /// Three test objects used, initialized in the constructor. Index 0 and 1 should be equal, while index 3
        /// should be different.
        /// </summary>
        protected DBPersistenceParameters[] testObject = new DBPersistenceParameters[3];

        /// <summary>
        /// Initializes <see cref="testObject"/>.
        /// </summary>
        public TestDBPersistenceParameters()
        {
            testObject[0] = new DBPersistenceParameters("test");
            testObject[1] = new DBPersistenceParameters("test");
            testObject[2] = new DBPersistenceParameters("test2");
        }

        /// <summary>
        /// Tests the <see cref="DBPersistenceParameters.ConnectionName"/> property.
        /// </summary>
        [Test]
        public void TestConnectionName()
        {
            Assert.AreEqual("test", testObject[0].ConnectionName, "Wrong name 0.");
            Assert.AreEqual("test", testObject[1].ConnectionName, "Wrong name 1.");
            // Don't check 2, so that can be used for objects that have "test" as connection name
            // in the subclass
        }

        /// <summary>
        /// Tests the <see cref="DBPersistenceParameters.GetHashCode"/> method. We assume that element 2 won't
        /// be a hash collision by pure chance.
        /// </summary>
        [Test]
        public void TestGetHashCode()
        {
            Assert.AreEqual(testObject[0].GetHashCode(), testObject[1].GetHashCode(),
                "0 and 1 should hash identically.");
            Assert.IsFalse(testObject[0].GetHashCode() == testObject[2].GetHashCode(),
                "0 and 2 should not hash identically.");
        }

        /// <summary>
        /// Tests equality among the three test objects.
        /// </summary>
        [Test]
        public void TestEquals()
        {
            Assert.IsTrue(testObject[0].Equals(testObject[1]), "0 == 1");
            Assert.IsTrue(testObject[1].Equals(testObject[0]), "1 == 0");
            Assert.IsFalse(testObject[0].Equals(testObject[2]), "0 != 2");
            Assert.IsFalse(testObject[2].Equals(testObject[1]), "2 != 1");
        }

        /// <summary>
        /// Equality with null is false.
        /// </summary>
        [Test]
        public void TestEqualsNull()
        {
            for (int i = 0; i < 3; i++)
            {
                Assert.IsFalse(testObject[i].Equals(null), String.Format("Equal to null: {0}", i));
            }
        }

        /// <summary>
        /// <c>testObject[0]</c> could be equal to a <see cref="DBDocumentIndexParameters"/> with the same connection
        /// name only if it is an <c>DBDocumentIndexParameters</c> itself.
        /// </summary>
        [Test]
        public void TestEqualsDocIndexParameters()
        {
            Assert.AreEqual(testObject[0].Equals(new DBDocumentIndexParameters("test", 0)),
                testObject[0] is DBDocumentIndexParameters,
                "Not equal to subclass.");
        }
    }

    /// <summary>
    /// This class tests the constructors of <see cref="DBPersistenceParameters"/>.
    /// </summary>
    [TestFixture]
    public class TestDBPersistenceParametersConstructors
    {
        /// <summary>
        /// Tests a null connection name to the constructor.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestConstructorNull()
        {
            new DBPersistenceParameters(null);
        }

        /// <summary>
        /// Tests an empty connection name to the constructor.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestConstructorEmpty()
        {
            new DBPersistenceParameters(" ");
        }
    }
}
