/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER
 * Version: 1.0
 */

using NUnit.Framework;
using System;
using System.Collections;
using System.IO;

namespace TopCoder.Document.Index.Persistence.Xml
{
    /// <summary>
    /// Tests the <see cref="XmlCollectionIndexSource"/> class. Most of the complex logic in that class relates to
    /// loading and saving XML representations.
    /// </summary>
    /// <remarks>
    /// This class "implements" <see cref="IParameters"/> to supply itself as an (incorrect) parameter object.
    /// </remarks>
    [TestFixture]
    public class TestXmlCollectionIndexSource : IParameters
    {
        /// <summary>
        /// A test object used in several tests, created based on a file.
        /// </summary>
        private XmlCollectionIndexSource testObject1;

        /// <summary>
        /// A test object crated with no file.
        /// </summary>
        private XmlCollectionIndexSource testObject2;

        /// <summary>
        /// An <see cref="IndexInfo"/> used for testing.
        /// </summary>
        private IndexInfo index;

        /// <summary>
        /// The name of the test file.
        /// </summary>
        private const string TEST_FILENAME = @"..\..\test_files\testcollections.xml";

        /// <summary>
        /// Changed files are saved under this name.
        /// </summary>
        private const string SAVE_FILENAME = SAVE_PATH + @"\savefile.xml";

        /// <summary>
        /// This is the path to a directory where files are saved.
        /// </summary>
        private const string SAVE_PATH = @"..\..\test_files\savedir\";

        /// <summary>
        /// Sets up the testing fixture.
        /// </summary>
        [SetUp]
        public void Setup()
        {
            if (!Directory.Exists(SAVE_PATH))
            {
                Directory.CreateDirectory(SAVE_PATH);
            }

            index = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlDocumentIndexSource",
                new SimpleIndexParameters("somefile.xml"));
            testObject1 = new XmlCollectionIndexSource();
            testObject1.Load(new SimpleIndexParameters(TEST_FILENAME));

            testObject2 = new XmlCollectionIndexSource();
        }

        /// <summary>
        /// Counts the elements reported by an enumerator.
        /// </summary>
        /// <param name="enumerator">The enumerator.</param>
        /// <returns>The number of elements.</returns>
        private static int CountEnumerator(IEnumerator enumerator)
        {
            int count = 0;
            while (enumerator.MoveNext())
            {
                count++;
            }

            return count;
        }

        /// <summary>
        /// Tests the <see cref="XmlCollectionIndexSource.FilePath"/> implementation. This property has both a
        /// getter and setter. The constructor-set value is verified, and then the value is changed.
        /// </summary>
        [Test]
        public void TestFilePath()
        {
            Assert.IsNull(testObject2.FilePath, "Null expected for never initialized path.");
            Assert.AreEqual(TEST_FILENAME, testObject1.FilePath, "Normal file path not returned correctly.");

            testObject2.FilePath = "newfilename";
            Assert.AreEqual("newfilename", testObject2.FilePath, "Changed file path not returned correctly.");
        }

        /// <summary>
        /// Tests setting a null file path. This is only allowed as the initial value, never to be set again.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestFilePathNull()
        {
            testObject1.FilePath = null;
        }

        /// <summary>
        /// Tests setting an empty file path. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestFilePathEmpty()
        {
            testObject1.FilePath = " ";
        }

        /// <summary>
        /// Tests setting an invalid file path, being a directory.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestFilePathDirectory()
        {
            testObject1.FilePath = @"../../test_files";
        }

        /// <summary>
        /// We check that we can successfully lock on the <see cref="XmlCollectionIndexSource.SyncRoot"/>, although
        /// that may not be the complete story.
        /// </summary>
        [Test]
        public void TestSyncRoot()
        {
            lock (testObject1.SyncRoot)
            {
            }
        }

        /// <summary>
        /// Tests constructing with a null filename.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestFileNameConstructorNull()
        {
            new XmlCollectionIndexSource(null);
        }

        /// <summary>
        /// Tests constructing with an empty filename.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestFileNameConstructorEmpty()
        {
            new XmlCollectionIndexSource(" ");
        }

        /// <summary>
        /// Tests constructing with an invalid filename.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestFileNameConstructorInvalid()
        {
            new XmlCollectionIndexSource("\0");
        }

        /// <summary>
        /// Tests the filename constructor.
        /// </summary>
        [Test]
        public void TestFileNameConstructor()
        {
            Assert.AreEqual("somefile.xml", new XmlCollectionIndexSource("somefile.xml").FilePath,
                "Reported path not as expected.");
        }

        /// <summary>
        /// Tests adding indices.
        /// </summary>
        [Test]
        public void TestAdd()
        {
            testObject1.Add("newindex", index);
            Assert.AreEqual(index, testObject1["newindex"], "The stored index should be returned.");

            testObject2.Add("index", index);
            Assert.AreEqual(index, testObject2["index"], "The stored index should be returned.");
        }

        /// <summary>
        /// The index info can store a reference to just about any class. However, we can only faithfully reproduce
        /// references to <see cref="XmlDocumentIndexSource"/>. Here, we reference an assembly which doesn't exist.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddInvalidAssembly()
        {
            IndexInfo index = new IndexInfo("TopCoder.Document.IDontExist.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlDocumentIndexSource",
                new SimpleIndexParameters("somefile.xml"));
            testObject1.Add("newindex", index);
        }

        /// <summary>
        /// The index info can store a reference to just about any class. However, we can only faithfully reproduce
        /// references to <see cref="XmlDocumentIndexSource"/>. Here, we reference the wrong class.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddWrongClass()
        {
            IndexInfo index = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlCollectionIndexSource",
                new SimpleIndexParameters("somefile.xml"));
            testObject1.Add("newindex", index);
        }

        /// <summary>
        /// Tests adding a null index. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestAddNull1()
        {
            testObject1.Add("newname", null);
        }

        /// <summary>
        /// Tests adding a null index name. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestAddNull2()
        {
            testObject1.Add(null, new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlDocumentIndexSource",
                new SimpleIndexParameters("somefile.xml")));
        }

        /// <summary>
        /// Tests adding an empty index name. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddEmpty()
        {
            testObject1.Add(" ", new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlDocumentIndexSource",
                new SimpleIndexParameters("somefile.xml")));
        }

        /// <summary>
        /// Tests removing an index, both one which really exists and one which does not.
        /// </summary>
        [Test]
        public void TestRemove()
        {
            testObject1.Remove("index");
            Assert.IsNull(testObject1["index"], "Index 'index' should now be null.");

            Assert.AreEqual(1, CountEnumerator(testObject1.GetEnumerator()), "Index count not as expected. 1");

            // The count should remain zero, but no exception shold be thrown
            testObject2.Remove("entre");
            Assert.IsNull(testObject2["entre"], "Index 'index' should still be null.");
            Assert.AreEqual(0, CountEnumerator(testObject2.GetEnumerator()), "Index count not as expected. 2");
        }

        /// <summary>
        /// Tests removing a null index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestRemoveNull()
        {
            testObject1.Remove(null);
        }

        /// <summary>
        /// Tests removing an empty name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestRemoveEmpty()
        {
            testObject1.Remove(" ");
        }

        /// <summary>
        /// A simple test of <see cref="XmlCollectionIndexSource.Contains"/>.
        /// </summary>
        [Test]
        public void TestContains()
        {
            Assert.IsTrue(testObject1.Contains("index"), "Index 'index' should exist.");
            Assert.IsFalse(testObject2.Contains("index"), "Index 'index' should not exist.");
        }

        /// <summary>
        /// Tests checking for a null index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestContainsNull()
        {
            testObject1.Contains(null);
        }

        /// <summary>
        /// Tests checking for an empty index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestContainsEmpty()
        {
            testObject1.Contains(" ");
        }

        /// <summary>
        /// Tests clearing the store.
        /// </summary>
        [Test]
        public void TestClear()
        {
            testObject1.Clear();
            Assert.IsNull(testObject1["index"], "Known index should now be null.");
            Assert.AreEqual(0, CountEnumerator(testObject1.GetEnumerator()), "Count should be zero after clearing.");
        }

        /// <summary>
        /// Tests <see cref="XmlCollectionIndexSource.GetEnumerator"/>, by iterating in order.
        /// </summary>
        [Test]
        public void TestGetEnumerator()
        {
            IDictionaryEnumerator enumerator = testObject1.GetEnumerator();
            for (int i = 0; i < 2; i++)
            {
                // Now, the specific order here WILL be dependent on the hashtable
                enumerator.MoveNext();
                Assert.AreEqual("index2", enumerator.Key, "The first key doesn't match.");
                Assert.AreEqual(@"..\..\test_files\filedoesntexist.xml",
                    (((IndexInfo) enumerator.Value).Parameters as SimpleIndexParameters).FileName,
                    "Wrong file name for first index.");

                enumerator.MoveNext();
                Assert.AreEqual("index", enumerator.Key, "The second key doesn't match.");
                Assert.AreEqual(@"..\..\test_files\testarchive.xml",
                    (((IndexInfo) enumerator.Value).Parameters as SimpleIndexParameters).FileName,
                    "Wrong file name for first index.");

                Assert.IsFalse(enumerator.MoveNext(), "Moving next should fail after end.");
                enumerator.Reset();
            }
        }

        /// <summary>
        /// Tests loading, both for the already loaded file, and the other one.
        /// </summary>
        [Test]
        public void TestLoadFiles()
        {
            testObject1.Load(new SimpleIndexParameters(TEST_FILENAME));
            // This test verifies that the data is correct quite thoroughly
            TestGetEnumerator();

            testObject1 = testObject2;
            // TestGetEnumerator is written for testObject1, so let's just move them around
            testObject1.Load(new SimpleIndexParameters(TEST_FILENAME));
            // This test verifies that the data is correct quite thoroughly
            TestGetEnumerator();
        }

        /// <summary>
        /// Tests loading with a null reference. Exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestLoadNull()
        {
            testObject1.Load(null);
        }

        /// <summary>
        /// Tests loading with an empty file name. Exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestLoadEmpty()
        {
            testObject1.Load(new SimpleIndexParameters(" "));
        }

        /// <summary>
        /// Tests loading with a parameter object of incorrect type.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestLoadWrongType()
        {
            // this implements IParameters
            testObject1.Load(this);
        }

        /// <summary>
        /// If there is no file name set, it's impossible to save.
        /// </summary>
        [Test]
        [ExpectedException(typeof(InvalidOperationException))]
        public void TestSaveInvalid()
        {
            testObject2.Save();
        }

        /// <summary>
        /// Tests saving over and over again in a subdirectory. The data is read back, and verification is also done to
        /// make sure that the total file count in the directory won't increase.
        /// </summary>
        [Test]
        public void TestSaving()
        {
            XmlCollectionIndexSource testObject = testObject1;
            testObject.FilePath = SAVE_FILENAME;
            for (int i = 0; i < 100; i++)
            {
                testObject.Save();

                // Save and load the data in another instance
                testObject1 = new XmlCollectionIndexSource();
                testObject1.Load(new SimpleIndexParameters(SAVE_FILENAME));
                TestGetEnumerator();
            }

            Assert.AreEqual(1, Directory.GetFiles(SAVE_PATH).Length, "Only one file should still be left.");
        }

        /// <summary>
        /// The getter part of the indexer is quite simple, we just verify that we retrieve correct data.
        /// </summary>
        [Test]
        public void TestGetIndexer()
        {
            IndexInfo index = testObject1["index2"];
            Assert.AreEqual(@"..\..\test_files\filedoesntexist.xml",
                    (index.Parameters as SimpleIndexParameters).FileName,
                    "Wrong file name.");

            Assert.IsNull(testObject1["I DON'T EXIST"],
                "Null should be returned for those words that are not present.");
        }

        /// <summary>
        /// Tests getting the list for a null index name.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestGetIndexerNull()
        {
            IndexInfo list = testObject1[null];
        }

        /// <summary>
        /// Tests getting the list for an empty index name.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestGetIndexerEmpty()
        {
            IndexInfo list = testObject1[" "];
        }

        /// <summary>
        /// Tests setting index infos with the indexer.
        /// </summary>
        [Test]
        public void TestSetIndexer()
        {
            testObject1["index"] = index;
            Assert.AreEqual(index, testObject1["index"], "The stored index should be returned.");

            testObject2["index"] = index;
            Assert.AreEqual(index, testObject2["index"], "The stored index should be returned.");
        }

        /// <summary>
        /// Tests setting the list for a null index name. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullIndexName()
        {
            testObject1[null] = index;
        }

        /// <summary>
        /// Tests setting the list for an empty index name. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestSetIndexerEmptyIndexName()
        {
            testObject1[" "] = index;
        }

        /// <summary>
        /// Tests setting a null index info. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullList()
        {
            testObject1["someindex"] = null;
        }

        /// <summary>
        /// Some tests save files for testing. Remove those files.
        /// </summary>
        [TearDown]
        public void Teardown()
        {
            foreach (string file in Directory.GetFiles(SAVE_PATH))
            {
                File.Delete(file);
            }
        }
    }
}
